searchNodes=[{"ref":"Exop.Chain.html","title":"Exop.Chain","type":"module","doc":"Provides macros to organize a number of Exop.Operation modules into an invocation chain. Example defmodule CreateUser do use Exop.Chain alias Operations.{User, Backoffice, Notifications} operation User.Create operation Backoffice.SaveStats operation Notifications.SendEmail end # CreateUser.run(name: &quot;User Name&quot;, age: 37, gender: &quot;m&quot;) Exop.Chain defines run/1 function that takes keyword() or map() of params. Those params will be passed into the first operation in the chain. Bear in mind that each of chained operations (except the first one) awaits a returned result of a previous operation as incoming params. So in the example above CreateUser.run(name: &quot;User Name&quot;, age: 37, gender: &quot;m&quot;) will invoke the chain by passing [name: &quot;User Name&quot;, age: 37, gender: &quot;m&quot;] params to the first User.Create operation. The result of User.Create operation will be passed to Backoffice.SaveStats operation as its params and so on. Once any of operations in the chain returns non-ok-tuple result (error result, interruption, auth error etc.) the chain execution interrupts and error result returned (as the chain (CreateUser) result)."},{"ref":"Exop.Chain.html#operation/2","title":"Exop.Chain.operation/2","type":"macro","doc":""},{"ref":"Exop.Chain.html#step/2","title":"Exop.Chain.step/2","type":"macro","doc":""},{"ref":"Exop.Fallback.html","title":"Exop.Fallback","type":"behaviour","doc":"Provides macros for fallback handling. Example defmodule FallbackModule do use Exop.Fallback def process(operation, params, error) do # your error handling code here :some_fallback_result end end defmodule TestOperation do use Exop.Operation fallback FallbackModule, return: true parameter :a, type: :integer parameter :b, type: :integer def process(%{a: a, b: b}), do: a + b end # TestOperation.run(a: 1, b: &quot;2&quot;) =&gt; :some_fallback_result By using Exop.Fallback you need to implement process/3 function which takes following params: failed operation module (TestOperation in the example above) params that were passed into the operation (%{a: 1, b: &quot;2&quot;}) an error result which was returned by the operation ({:error, {:validation, %{a: [&quot;has wrong type&quot;]}}}) During a fallback definition you can add return: true option so in the example case TestOperation.run/1 will return the result of the fallback (FallbackModule.process/3 function&#39;s result - :some_fallback_result). If you want TestOperation.run/1 to return original result ({:error, {:validation, %{a: [&quot;has wrong type&quot;]}}}) specify return: false option or just omit it in a fallback definition."},{"ref":"Exop.Fallback.html#c:process/3","title":"Exop.Fallback.process/3","type":"callback","doc":"Fallback handling function. Receives: failed operation module a map of parameters with which the operation was invoked an error tuple returned by the operation"},{"ref":"Exop.Fallback.html#t:t/0","title":"Exop.Fallback.t/0","type":"type","doc":""},{"ref":"Exop.Operation.html","title":"Exop.Operation","type":"behaviour","doc":"Provides macros for an operation&#39;s contract definition and process/1 function. Example defmodule SomeOperation do use Exop.Operation parameter :param1, type: :integer, required: true parameter :param2, type: :string, length: %{max: 3}, format: ~r/foo/ def process(params) do &quot;This is the operation&#39;s result with one of the params = &quot; &lt;&gt; params[:param1] end end"},{"ref":"Exop.Operation.html#authorize/1","title":"Exop.Operation.authorize/1","type":"macro","doc":"Authorizes an action with predefined policy (see policy macro docs). If authorization fails, any code after (below) auth check will be postponed (an error {:error, {:auth, _reason}} will be returned immediately)"},{"ref":"Exop.Operation.html#current_policy/0","title":"Exop.Operation.current_policy/0","type":"macro","doc":"Returns policy that was defined in an operation."},{"ref":"Exop.Operation.html#fallback/2","title":"Exop.Operation.fallback/2","type":"macro","doc":"Defines a fallback module that will be used for an operation&#39;s non-ok-tuple (fail) result handling. defmodule MultiplyByTenOperation do use Exop.Operation fallback LoggerFallback parameter :a, type: :integer, required: true def process(%{a: a}), do: a * 10 end A fallback module itself might be: defmodule LoggerFallback do use Exop.Fallback require Logger def process(operation_module, params_passed_to_the_operation, operation_error_result) do Logger.error(&quot;Oops&quot;) end end If return: true option is provided then failed operation&#39;s run/1 will return the fallback&#39;s process/3 result."},{"ref":"Exop.Operation.html#parameter/2","title":"Exop.Operation.parameter/2","type":"macro","doc":"Defines a parameter with name and opts in an operation contract. Options could include the parameter value checks and transformations (like coercion). Example parameter :some_param, type: :map, required: true Available checks are: type Checks whether a parameter&#39;s value is of declared type. parameter :some_param, type: :map required Checks the presence of a parameter in passed params collection. parameter :some_param, required: true default Checks if the parameter is missed and assigns default value to it if so. parameter :some_param, default: &quot;default value&quot; numericality Checks whether a parameter&#39;s value is a number and passes constraints (if constraints were defined). parameter :some_param, numericality: %{equal_to: 10, greater_than: 0, greater_than_or_equal_to: 10, less_than: 20, less_than_or_equal_to: 10} in Checks whether a parameter&#39;s value is within a given list. parameter :some_param, in: ~w(a b c) not_in Checks whether a parameter&#39;s value is not within a given list. parameter :some_param, not_in: ~w(a b c) format Checks wether parameter&#39;s value matches given regex. parameter :some_param, format: ~r/foo/ length Checks the length of a parameter&#39;s value. parameter :some_param, length: %{min: 5, max: 10, is: 7, in: 5..8} inner Checks the inner of either Map or Keyword parameter. parameter :some_param, type: :map, inner: %{ a: [type: :integer, required: true], b: [type: :string, length: %{min: 1, max: 6}] } struct Checks whether the given parameter is expected structure. parameter :some_param, struct: %SomeStruct{} list_item Checks whether each of list items conforms defined checks. An item&#39;s checks could be any that Exop offers: parameter :list_param, list_item: %{type: :string, length: %{min: 7}} func Checks whether an item is valid over custom validation function. parameter :some_param, func: &amp;__MODULE__.your_validation/2 def your_validation(_params, param), do: !is_nil(param) allow_nil It is not a parameter check itself, because it doesn&#39;t return any validation errors. It is a parameter attribute which allow you to have other checks for a parameter whilst have a possibility to pass nil as the parameter&#39;s value. If nil is passed all the parameter&#39;s checks are ignored during validation. Coercion It is possible to coerce a parameter before the contract validation, all validation checks will be invoked on coerced parameter value. Since coercion changes a parameter before any validation has been invoked, default values are resolved (with :default option) before the coercion. The flow looks like: Resolve param default value -&gt; Coerce -&gt; Validate coerced parameter :some_param, default: 1, numericality: %{greater_than: 0}, coerce_with: &amp;__MODULE__.coerce/1 def coerce(x), do: x * 2 For more information and examples check out general Exop docs."},{"ref":"Exop.Operation.html#policy/2","title":"Exop.Operation.policy/2","type":"macro","doc":"Defines a policy that will be used for authorizing the possibility of a user to invoke an operation. defmodule ReadOperation do use Exop.Operation policy MonthlyReportPolicy, :can_read? parameter :user, required: true, struct: %User{} def process(params) do authorize(params.user) # make some reading... end end A policy itself might be: defmodule MonthlyReportPolicy do # not only Keyword or Map as an argument since 1.1.1 def can_read?(%User{role: &quot;manager&quot;}), do: true def can_read?(_opts), do: false def can_write?(%User{role: &quot;manager&quot;}), do: true def can_write?(_opts), do: false end"},{"ref":"Exop.Operation.html#c:process/1","title":"Exop.Operation.process/1","type":"callback","doc":"Operation&#39;s entry point. Takes defined contract as the single parameter. Contract itself is a list of maps: [%{name: atom(), opts: keyword()}]"},{"ref":"Exop.Policy.html","title":"Exop.Policy","type":"behaviour","doc":"Provides macros for policy validation. Example defmodule MonthlyReportPolicy do # not only Keyword or Map as an argument since 1.1.1 def can_read?(%{user_role: &quot;admin&quot;}), do: true def can_read?(&quot;admin&quot;), do: true def can_read?(%User{role: &quot;manager&quot;}), do: true def can_read?(:manager), do: true def can_read?(_opts), do: false def can_write?(%{user_role: &quot;manager&quot;}), do: true def can_write?(_opts), do: false end defmodule ReadOperation do use Exop.Operation policy MonthlyReportPolicy, :can_read? parameter :user, required: true, struct: %User{} def process(params) do authorize(params.user) # make some reading... end end If authorization fails, any code after (below) auth check is postponed: an error {:error, {:auth, _reason}} is returned immediately."},{"ref":"Exop.Policy.html#c:authorize/2","title":"Exop.Policy.authorize/2","type":"callback","doc":"Authorizes the possibility to invoke an action."},{"ref":"Exop.Policy.html#t:t/0","title":"Exop.Policy.t/0","type":"type","doc":""},{"ref":"Exop.StringCharsImplementations.html","title":"Exop.StringCharsImplementations","type":"module","doc":"String.Chars protocol implementations for some types"},{"ref":"Exop.TypeValidation.html","title":"Exop.TypeValidation","type":"module","doc":""},{"ref":"Exop.TypeValidation.html#check_value/2","title":"Exop.TypeValidation.check_value/2","type":"function","doc":""},{"ref":"Exop.TypeValidation.html#known_types/0","title":"Exop.TypeValidation.known_types/0","type":"function","doc":""},{"ref":"Exop.TypeValidation.html#type_supported?/1","title":"Exop.TypeValidation.type_supported?/1","type":"function","doc":""},{"ref":"Exop.Utils.html","title":"Exop.Utils","type":"module","doc":"A bunch of common functions."},{"ref":"Exop.Utils.html#try_map/1","title":"Exop.Utils.try_map/1","type":"function","doc":"Tries to make a map from a struct and keyword list"},{"ref":"Exop.Validation.html","title":"Exop.Validation","type":"module","doc":"Provides high-level functions for a contract validation. The main function is valid?/2 Mostly invokes Exop.ValidationChecks module functions."},{"ref":"Exop.Validation.html#check_inner/3","title":"Exop.Validation.check_inner/3","type":"function","doc":"Checks inner item of the contract param (which is a Map itself) with their own checks. Examples iex&gt; Exop.Validation.check_inner(%{a: 1}, :a, [b: [type: :atom], c: [type: :string]]) [%{a: &quot;has wrong type&quot;}] iex&gt; Exop.Validation.check_inner(%{a: []}, :a, [b: [type: :atom], c: [type: :string]]) [[%{&quot;a[:b]&quot; =&gt; &quot;is required&quot;}, true], [%{&quot;a[:c]&quot; =&gt; &quot;is required&quot;}, true]] iex&gt; Exop.Validation.check_inner(%{a: %{b: :atom, c: &quot;string&quot;}}, :a, [b: [type: :atom], c: [type: :string]]) [[true, true], [true, true]]"},{"ref":"Exop.Validation.html#check_list_item/3","title":"Exop.Validation.check_list_item/3","type":"function","doc":"Checks items of the contract&#39;s list param with specified checks. Examples iex&gt; Exop.Validation.check_list_item(%{a: 1}, :a, [type: :integer]) [%{a: &quot;is not a list&quot;}] iex&gt; Exop.Validation.check_list_item(%{a: []}, :a, [type: :integer]) [] iex&gt; Exop.Validation.check_list_item(%{a: [1, :atom]}, :a, [type: :integer]) [[true, true], [true, %{&quot;a[1]&quot; =&gt; &quot;has wrong type&quot;}]] iex&gt; Exop.Validation.check_list_item(%{a: [1, 2]}, :a, [type: :integer]) [[true, true], [true, true]]"},{"ref":"Exop.Validation.html#errors_message/1","title":"Exop.Validation.errors_message/1","type":"function","doc":""},{"ref":"Exop.Validation.html#valid?/2","title":"Exop.Validation.valid?/2","type":"function","doc":"Validate received params over a contract. Examples iex&gt; Exop.Validation.valid?([%{name: :param, opts: [required: true]}], [param: &quot;hello&quot;]) :ok"},{"ref":"Exop.Validation.html#validate/3","title":"Exop.Validation.validate/3","type":"function","doc":"Validate received params over a contract. Accumulate validation results into a list. Examples iex&gt; Exop.Validation.validate([%{name: :param, opts: [required: true, type: :string]}], [param: &quot;hello&quot;], []) [true, true]"},{"ref":"Exop.Validation.html#t:validation_error/0","title":"Exop.Validation.validation_error/0","type":"type","doc":""},{"ref":"Exop.Validation.ValidationError.html","title":"Exop.Validation.ValidationError","type":"exception","doc":"An operation&#39;s contract validation failure error."},{"ref":"Exop.ValidationChecks.html","title":"Exop.ValidationChecks","type":"module","doc":"Provides low-level validation functions: check_type/3 check_required/3 check_numericality/3 check_in/3 check_not_in/3 check_format/3 check_length/3 check_struct/3 check_func/3 check_equals/3 check_exactly/3 check_allow_nil/3"},{"ref":"Exop.ValidationChecks.html#check_allow_nil/3","title":"Exop.ValidationChecks.check_allow_nil/3","type":"function","doc":""},{"ref":"Exop.ValidationChecks.html#check_equals/3","title":"Exop.ValidationChecks.check_equals/3","type":"function","doc":"Checks whether a parameter&#39;s value exactly equals given value (with type equality). Examples iex&gt; Exop.ValidationChecks.check_equals(%{a: 1}, :a, 1) true"},{"ref":"Exop.ValidationChecks.html#check_exactly/3","title":"Exop.ValidationChecks.check_exactly/3","type":"function","doc":"The alias for check_equals/3. Checks whether a parameter&#39;s value exactly equals given value (with type equality). Examples iex&gt; Exop.ValidationChecks.check_exactly(%{a: 1}, :a, 1) true"},{"ref":"Exop.ValidationChecks.html#check_format/3","title":"Exop.ValidationChecks.check_format/3","type":"function","doc":"Checks whether an item_name conforms the given format. Examples iex&gt; Exop.ValidationChecks.check_format(%{a: &quot;bar&quot;}, :a, ~r/bar/) true"},{"ref":"Exop.ValidationChecks.html#check_func/3","title":"Exop.ValidationChecks.check_func/3","type":"function","doc":"Checks whether an item is valid over custom validation function. Examples iex&gt; Exop.ValidationChecks.check_func(%{a: 1}, :a, fn(_contract, param)-&gt; param &gt; 0 end) true iex&gt; Exop.ValidationChecks.check_func(%{a: 1}, :a, fn(_contract, param)-&gt; is_nil(param) end) %{a: &quot;isn&#39;t valid&quot;} iex&gt; Exop.ValidationChecks.check_func(%{a: -1}, :a, fn(_contract, _param)-&gt; {:error, :my_error} end) %{a: :my_error} iex&gt; Exop.ValidationChecks.check_func(%{a: -1, b: 1}, :a, fn(_contract, param_name, param_value)-&gt; {param_name, param_value} == {:a, -1} end) true iex&gt; Exop.ValidationChecks.check_func(%{a: -1, b: 1}, :b, fn(_contract, param_name, param_value)-&gt; {param_name, param_value} != {:a, -1} end) true"},{"ref":"Exop.ValidationChecks.html#check_in/3","title":"Exop.ValidationChecks.check_in/3","type":"function","doc":"Checks whether an item_name is a memeber of a list. Examples iex&gt; Exop.ValidationChecks.check_in(%{a: 1}, :a, [1, 2, 3]) true"},{"ref":"Exop.ValidationChecks.html#check_item_present?/2","title":"Exop.ValidationChecks.check_item_present?/2","type":"function","doc":"Checks whether a check_item has been provided. Returns a boolean. Examples iex&gt; Exop.ValidationChecks.check_item_present?(%{a: 1, b: 2}, :a) true iex&gt; Exop.ValidationChecks.check_item_present?([a: 1, b: 2], :b) true iex&gt; Exop.ValidationChecks.check_item_present?(%{a: 1, b: 2}, :c) false iex&gt; Exop.ValidationChecks.check_item_present?(%{a: 1, b: nil}, :b) true"},{"ref":"Exop.ValidationChecks.html#check_length/3","title":"Exop.ValidationChecks.check_length/3","type":"function","doc":"Checks an item_name over length constraints. Examples iex&gt; Exop.ValidationChecks.check_length(%{a: &quot;123&quot;}, :a, %{min: 0}) [true] iex&gt; Exop.ValidationChecks.check_length(%{a: ~w(1 2 3)}, :a, %{in: 2..4}) [true] iex&gt; Exop.ValidationChecks.check_length(%{a: ~w(1 2 3)}, :a, %{is: 3, max: 4}) [true, true]"},{"ref":"Exop.ValidationChecks.html#check_not_in/3","title":"Exop.ValidationChecks.check_not_in/3","type":"function","doc":"Checks whether an item_name is not a memeber of a list. Examples iex&gt; Exop.ValidationChecks.check_not_in(%{a: 4}, :a, [1, 2, 3]) true"},{"ref":"Exop.ValidationChecks.html#check_numericality/3","title":"Exop.ValidationChecks.check_numericality/3","type":"function","doc":"Checks an item_name over numericality constraints. Examples iex&gt; Exop.ValidationChecks.check_numericality(%{a: 3}, :a, %{ equal_to: 3 }) true iex&gt; Exop.ValidationChecks.check_numericality(%{a: 5}, :a, %{ greater_than_or_equal_to: 3 }) true iex&gt; Exop.ValidationChecks.check_numericality(%{a: 3}, :a, %{ less_than_or_equal_to: 3 }) true"},{"ref":"Exop.ValidationChecks.html#check_regex/3","title":"Exop.ValidationChecks.check_regex/3","type":"function","doc":"The alias for check_format/3. Checks whether an item_name conforms the given format. Examples iex&gt; Exop.ValidationChecks.check_regex(%{a: &quot;bar&quot;}, :a, ~r/bar/) true"},{"ref":"Exop.ValidationChecks.html#check_required/3","title":"Exop.ValidationChecks.check_required/3","type":"function","doc":"Checks if an item_name presents in params if its required (true). Examples iex&gt; Exop.ValidationChecks.check_required(%{}, :some_item, false) true iex&gt; Exop.ValidationChecks.check_required([a: 1, b: 2], :a, true) true iex&gt; Exop.ValidationChecks.check_required(%{a: 1, b: 2}, :b, true) true"},{"ref":"Exop.ValidationChecks.html#check_struct/3","title":"Exop.ValidationChecks.check_struct/3","type":"function","doc":"Checks whether an item is expected structure. Examples defmodule SomeStruct1, do: defstruct [:a, :b] defmodule SomeStruct2, do: defstruct [:b, :c] Exop.ValidationChecks.check_struct(%{a: %SomeStruct1{}}, :a, %SomeStruct1{}) # true Exop.ValidationChecks.check_struct(%{a: %SomeStruct1{}}, :a, %SomeStruct2{}) # false"},{"ref":"Exop.ValidationChecks.html#check_type/3","title":"Exop.ValidationChecks.check_type/3","type":"function","doc":"Checks the type of an item_name. Examples iex&gt; Exop.ValidationChecks.check_type(%{a: 1}, :a, :integer) true iex&gt; Exop.ValidationChecks.check_type(%{a: &quot;1&quot;}, :a, :string) true iex&gt; Exop.ValidationChecks.check_type(%{a: nil}, :a, :string) %{:a =&gt; &quot;has wrong type&quot;}"},{"ref":"Exop.ValidationChecks.html#get_check_item/2","title":"Exop.ValidationChecks.get_check_item/2","type":"function","doc":"Returns an check_item&#39;s value from either a Keyword or a Map by an atom-key. Examples iex&gt; Exop.ValidationChecks.get_check_item(%{a: 1, b: 2}, :a) 1 iex&gt; Exop.ValidationChecks.get_check_item([a: 1, b: 2], :b) 2 iex&gt; Exop.ValidationChecks.get_check_item(%{a: 1, b: 2}, :c) nil"},{"ref":"Exop.ValidationChecks.html#t:check_error/0","title":"Exop.ValidationChecks.check_error/0","type":"type","doc":""},{"ref":"readme.html","title":"Exop","type":"extras","doc":"Exop A library that helps you to organize your Elixir code in more domain-driven way. Exop provides macros which helps you to encapsulate business logic and offers you additionally: incoming params validation (with predefined contract), params coercion, policy check, fallback behavior and more. ExopData Interested in property-based testing? Check out new Exop family member - ExopData. If you use Exop to organize your code with ExopData you can get property generators in the most easiest way."},{"ref":"readme.html#table-of-contents","title":"Exop - Table of Contents","type":"extras","doc":"Here is the CHANGELOG that was started from ver. 0.4.1 ¯\\_(ツ)_/¯ Installation Operation definition Parameter checks type required default numericality equals in not_in format length inner struct list_item func allow_nil from Defined params Interrupt Coercion Policy check Fallback module Operation invocation Operation results Operations chain"},{"ref":"readme.html#installation","title":"Exop - Installation","type":"extras","doc":"def deps do [{:exop, &quot;~&gt; 1.3.0&quot;}] end"},{"ref":"readme.html#operation-definition","title":"Exop - Operation definition","type":"extras","doc":"defmodule IntegersDivision do use Exop.Operation parameter :a, type: :integer, default: 1 parameter :b, type: :integer, required: false, numericality: %{greater_than: 0} def process(params) do result = params[:a] / params[:b] IO.inspect &quot;The division result is: \#{result}&quot; end end Exop.Operation provides parameter macro, which is responsible for the contract definition. Its spec is @spec parameter(atom | String.t, Keyword.t) :: none, we define parameter name as the first argument and parameter options as the second Keyword argument. A parameter name could be either an atom or a string. You could even mix atom-named and string-named parameters in an operation&#39;s contract. Parameter options determine a contract of a parameter, a set of parameters contracts is an operation contract. Business logic of an operation is defined in process/1 function, which is required by the Exop.Operation module behaviour. After the contract and business logic were defined, you can invoke the operation simply by calling run/1 function: iex&gt; IntegersDivision.run(a: 50, b: 5) {:ok, &quot;The division result is: 10&quot;} Return type will be either {:ok, any()} (where the second item in the tuple is process/1 function&#39;s result) or {:error, {:validation, map()}} (where the map() is validation errors map). for more information see Operation results section Parameter checks A parameter options could have various checks. Here the list of checks available yet: type required default numericality equals (exactly) in not_in format (regex) length inner struct list_item func allow_nil type Checks whether a parameter&#39;s value is of declared type. parameter :some_param, type: :map Exop handle almost all Elixir types and some additional: :boolean :integer :float :string :tuple :map :keyword :list :atom :module :function :uuid Unknown type always generates ArgumentError exception on compile time. module &#39;type&#39; means Exop expects a parameter&#39;s value to be an atom (a module name) and this module should be already loaded (ready to call it&#39;s functions) uuid is not actually a &quot;type&quot; but I placed this under :type check because there is no reason to have dedicated :uuid check. required Checks the presence/absence of a parameter in passed to run/1 params collection. Given parameters collection fails the validation only if required parameter is missed, if required parameter&#39;s value is nil this parameter will pass this check. parameter :param_a # the same as required: true, required by default parameter :param_b, required: false # this parameter is not required By default, a parameter is required (since version 1.2.0, required: true). If you want to specify a parameter is not required, provide required: false. Why? Because you might find that you repetitively type required: true for almost every parameter in a contract. I think if you provide a parameter to an operation (define it in a contract) you expect to get it. Cases, when you need a parameter passed into an operation (and don&#39;t really care whether it is present or not), are pretty rare. Since version 1.1.0 the behavior of this check has been changed. Check out CHANGELOG for more info. default Checks the presence of a parameter in passed to run/1 params collection, and if the parameter is missed - assign default value to it. parameter :some_param, default: &quot;default value&quot; numericality Checks whether a parameter&#39;s value is a number and other numeric constraints. All possible constraints are listed in the example below. parameter :some_param, numericality: %{equal_to: 10, # (aliases: `equals`, `is`, `eq`) greater_than: 0, # (alias: `gt`) greater_than_or_equal_to: 10 # (aliases: `min`, `gte`), less_than: 20, # (alias: `lt`) less_than_or_equal_to: 10 # (aliases: `max`, `lte`)} equals (alias: exactly) Checks whether a parameter&#39;s value exactly equals given value (with type equality). parameter :some_param, equals: 100.5 parameter :some_param, exactly: 100.5 in Checks whether a parameter&#39;s value is within a given list. parameter :some_param, in: ~w(a b c) not_in Checks whether a parameter&#39;s value is not within a given list. parameter :some_param, not_in: ~w(a b c) format (alias: regex) Checks wether parameter&#39;s value matches given regex. parameter :some_param, format: ~r/foo/ parameter :some_param, regex: ~r/foo/ length Checks the length of a parameter&#39;s value. The value should be one of handled types: list (items count) string (chars count) atom (treated as string) map (key-value pairs count) tuple (items count) length check is complex as numericality (should define map of inner checks). All possible checks are listed in the example below. parameter :some_param, length: %{gte: 5, gt: 4, min: 5, lte: 10, lt: 9, max: 10, is: 7, in: 5..8} inner Checks the inner of either Map or Keyword parameter. It applies checks described in inner map to related inner items. # some_param = %{a: 3, b: &quot;inner_b_attr&quot;} parameter :some_param, type: :map, inner: %{ a: [type: :integer], b: [type: :string, length: %{min: 1, max: 6}] } # you can omit `type` and `inner` checks keywords in order to check inner of your parameter, # when `type` hasn&#39;t been specified explicitly, both keyword and map types pass the `type` validation parameter :some_param, %{ a: [type: :integer], b: [type: :string, length: %{min: 1, max: 6}] } And, of course, all checks on a parent parameter (:some_param in the example) are still applied. struct Checks whether the given parameter is expected structure. parameter :some_param, struct: %SomeStruct{} # or parameter :some_param, struct: SomeStruct list_item Checks whether each of list items conforms defined checks. An item&#39;s checks could be any that Exop offers: # list_param = [&quot;1234567&quot;, &quot;7chars&quot;] # you can omit `type` check while you&#39;re passing a list to an operation parameter :list_param, list_item: %{type: :string, length: %{min: 7}} Even more complex like inner: # list_param = [ # %TestStruct{a: 3, b: &quot;6chars&quot;}, # %TestStruct{a: nil, b: &quot;7charss&quot;} # ] parameter :list_param, list_item: %{inner: %{ a: %{type: :integer}, b: %{type: :string, length: %{min: 7}} }} Moreover, coerce_with and default options are available too. func Checks whether an item is valid over custom validation function. If this function returns false, validation will fail with default message &quot;isn&#39;t valid&quot;. # arity of 1 parameter :some_param, func: &amp;is_atom/1 # arity of 2 parameter :some_param, func: &amp;__MODULE__.your_validation/2 def your_validation(_params, param_value), do: !is_nil(param_value) # arity of 3 parameter :some_param, func: &amp;__MODULE__.your_validation/3 def your_validation(_params, :some_param = _param_name, param_value), do: !is_nil(param_value) A custom validation function can also return a user-specified message which will be displayed in map of validation errors. def your_validation(_params, param) do if param &gt; 99 do true else {:error, &quot;Custom error message&quot;} end end Therefore, validation will fail, if the function returns either false or {:error, your_error_msg} tuple. func/1 receives one argument which is the actual parameter value to check. func/2 receives two arguments: the first is a contract of an operation (parameters with their values), the second - the actual parameter value to check. So, now you can validate a parameter depending on other parameters values. parameter :a, type: :integer parameter :b, func: &amp;__MODULE__.your_validation/2 def your_validation(params, b), do: params[:a] &gt; 0 &amp;&amp; !is_nil(b) it&#39;s possible to combine :func check with others (though not preferable), just make sure this check is the last check in the list allow_nil It is not a parameter check itself, because it doesn&#39;t return any validation errors. It is a parameter attribute which allow you to have other checks for a parameter whilst have a possibility to pass nil as the parameter&#39;s value. If nil is passed all the parameter&#39;s checks are ignored during validation. defmodule YourOperation do use Exop.Operation parameter :a, type: :integer, allow_nil: true parameter :b, type: :integer, allow_nil: false def process(params), do: params end {:ok, %{a: 1}} = YourOperation.run(a: 1, b: 1) {:ok, %{a: nil}} = YourOperation.run(a: nil, b: 1) {:ok, %{b: 1}} = YourOperation.run(a: nil, b: 1) {:error, {:validation, %{b: [&quot;doesn&#39;t allow nil&quot;, &quot;has wrong type&quot;]}}} = YourOperation.run(a: nil, b: nil) By default (if you omit allow_nil attribute), a parameter is treated as allow_nil: false from This option allows you to pass a parameter to run/1 and run!/1 functions with one name and work with this parameter within an operation under another name. defmodule YourOperation do use Exop.Operation parameter :a, type: :integer, from: &quot;a&quot; parameter :b, type: :string, from: :bB def process(params), do: params end # now you can invoke YourOperation with such params: # Youroperation.run(%{&quot;a&quot; =&gt; 1, b: &quot;1&quot;}) # Youroperation.run(%{a: 1, bB: &quot;1&quot;}) # Youroperation.run(%{&quot;a&quot; =&gt; 1, bB: &quot;1&quot;}) # and get: # {:ok, %{a: 1, b: &quot;1&quot;}} The same works for parameters given as a Keyword as well (in this case :from value should be an atom): defmodule YourOperation do use Exop.Operation parameter :a, type: :integer, from: :aA parameter :b, type: :string, from: :bB def process(params), do: params end # Youroperation.run(a: 1, b: &quot;1&quot;) # Youroperation.run(aA: 1, bB: &quot;1&quot;) # {:ok, %{a: 1, b: &quot;1&quot;}} Why? Simply because sometimes you&#39;re not in control of incoming parameters but don&#39;t want to map them each time you need to use&#39;em by yourself (good example: params in Phoenix controller&#39;s action, which come as a map with string keys). This option doesn&#39;t work for :inner check&#39;s inner parameters currently. Interrupt In some cases you might want to make an &#39;early return&#39; from process/1 function. For this purpose you can call interrupt/1 function within process/1 and pass an interruption reason to it. An operation will be interrupted and return {:interrupt, your_reason} # ... def process(_params) do interrupt(%{fail: &quot;oops&quot;}) :ok # will not return it end # ... SomeOperation.run(a: 1) # {:interrupt, %{fail: &quot;oops&quot;}} run!/1 invocation doesn&#39;t affect interruption result: {:interrupt, _your_result} tuple will be returned anyway as expected and handled result. Coercion It is possible to coerce a parameter before the contract validation, all validation checks will be invoked on coerced parameter value. Since coercion changes a parameter before any validation has been invoked, default values are resolved (with :default option) before the coercion. The flow looks like: Resolve param default value -&gt; Coerce -&gt; Validate coerced If coercion function returns an error tuple it will be treated as validation failure: an operation&#39;s invokation stops and that error tuple will be returned as a result. :coerce_with accepts 2-arity function. This function takes a tuple {coerced_param_name, coerced_param_value} as the first argument and a map with all the parameters that have been passed to either run/1 or run!/1 function as the second argument. parameter :a, type: :integer parameter :b, type: :string, coerce_with: &amp;__MODULE__.to_string/2 parameter :c, type: :string, coerce_with: &amp;__MODULE__.to_string/2 def to_string({_, value}, %{a: _, b: _, c: _} = _received_params) when is_integer(value) do Integer.to_string(c_value) end def to_string({_, value}, _received_params) when is_binary(c_value) do value end def to_string({:c, c_value}, _received_params) do # special coercion for :c parameter here end Why it is so? Because there are cases when you can use the same coercion function for multiple params and/or coerce a parameter depending on another&#39;s value and need to have all this information. Policy check It is possible to define a policy that will be used for authorizing the possibility to invoke an operation. So far, there is a simple policy implementation and usage: first of all, define a policy module (use Exop.Policy is not actual since ver. 1.1.1 - it is not mandatory to use this macro. Just define a module with a bunch of functions that take a single argument (any type) and return either true or false) defmodule MonthlyReportPolicy do # not only Keyword or Map as an argument since 1.1.1 def can_read?(%{user_role: &quot;admin&quot;}), do: true def can_read?(&quot;admin&quot;), do: true def can_read?(%User{role: &quot;manager&quot;}), do: true def can_read?(:manager), do: true def can_read?(_opts), do: false def can_write?(%{user_role: &quot;manager&quot;}), do: true def can_write?(_opts), do: false end In this policy two actions (checks) defined (can_read?/1 &amp; can_write?/1). Every action expects an argument for a check. It&#39;s up to you how to handle this argument and turn it into the actual check. Bear in mind: only true return-value treated as true, everything else returned form an action treated as false next step - link an operation and a policy defmodule ReadOperation do use Exop.Operation policy MonthlyReportPolicy, :can_read? parameter :user, struct: %User{} def process(%{user: %User{} = user}) do # make some reading... end end finally - call authorize/1 within process/1 defmodule ReadOperation do use Exop.Operation policy MonthlyReportPolicy, :can_read? parameter :user, struct: %User{} def process(params) do authorize(params.user) # make some reading... end end Please, note: if authorization fails, any code after (below) auth check will be postponed (an error {:error, {:auth, _policy_action_name}} will be returned immediately as an operation result) Fallback module If you&#39;d like to handle various operations fails with a certain logic (for example log it into Graylog) you can use Exop.Fallback. Define a fallback module: defmodule FallbackModule do use Exop.Fallback def process(operation, params, error) do # your error handling code here :some_fallback_result end end here you need to define and implement process/3 function which takes following params: failed operation module params that were passed into the operation an error result which was returned by the operation Use your fallback in operations like this: defmodule SomeOperation do use Exop.Operation fallback FallbackModule, return: true parameter :a, type: :integer parameter :b, type: :integer def process(%{a: a, b: b}), do: a + b end The results of the operation executions: # SomeOperation will fail iex&gt; SomeOperation.run(a: 1, b: &quot;2&quot;) :some_fallback_result # SomeOperation will be successful iex&gt; SomeOperation.run(a: 1, b: 2) {:ok, 3} During a fallback definition you can add return: true option so in the example case SomeOperation.run/1 will return the result of the fallback (FallbackModule.process/3 function&#39;s result - :some_fallback_result). If you want SomeOperation.run/1 to return original result ({:error, {:validation, %{a: [&quot;has wrong type&quot;]}}}) specify return: false option or just omit it in a fallback definition."},{"ref":"readme.html#operation-invocation","title":"Exop - Operation invocation","type":"extras","doc":"As said earlier, operations in most cases called by run/1 function. This function receives parameters collection. It&#39;s not required to pass to run/1 function parameters only described in the operation&#39;s contract, but only described parameters will be validated and gived to process/1 function, all other will be filtered out from the process. filtering out parameters which are not defined in a contract is here to support the whole idea of defining the right operation&#39;s contract and take care of what you really need for a certain business process / function run/1 function validates received parameters over the contract and if all parameters passed the validation, the run/1 function calls the code defined in process/1 function. iex&gt; SomeOperation.run(param1: 1, param2: &quot;2&quot;) _some_result_ If at least one of the given parameters didn&#39;t pass the validation process/1 function&#39;s code will not be invoked and corresponding warning in the application&#39;s log will appear. There is &quot;bang&quot; version of run/1 exists. Function run!/1 does the same things that its sibling does, the only difference is a result of invocation, it might be: if a contract validation passed - the actual result of an operation (result of a code, described in process/1) if a contract validation failed - an error Exop.Validation.ValidationError raising in case of manual interruption - {:interrupt, _reason} You always can bypass the validation simply by calling process/1 function itself, if needed."},{"ref":"readme.html#operation-results","title":"Exop - Operation results","type":"extras","doc":"If received parameters passed a contract validation, a code defined in process/1 will be invoked. Or you will receive @type validation_error :: {:error, :validation_failed, map()} as a result otherwise. map() as errors reasons might look like this: %{param1: [&quot;has wrong type&quot;], param2: [&quot;is required&quot;, &quot;must be equal to 3&quot;]} An operation can return one of results listed below (depends on passed in params and operation definition): an operation was completed successfully: {:error, _your_error_reason_} (if an :error-tuple (any length, but :error atom should be the first element) was returned by process/1 function) {:ok, any()} (otherwise, even if {:ok, _your_result_} tuple was returned by process/1 function) a contract validation failed: {:error, {:validation, map()}} if interrupt/1 was invoked: {:interrupt, any()} policy check failed: {:error, {:auth, :undefined_policy}} {:error, {:auth, :undefined_action}} {:error, {:auth, atom()}} For the &quot;bang&quot; version of run/1 see results description above."},{"ref":"readme.html#operations-chain","title":"Exop - Operations chain","type":"extras","doc":"Sometimes you need to aggregate/group &#39;atom&#39; operations into a single one operation responsible for some complex business process/logic. You have a few approaches to do it (with for example) but mb you&#39;ll find Exop.Chain more handy. Exop.Chain provides a simple way to organize a number of Exop.Operation modules into an invocation chain. defmodule CreateUser do use Exop.Chain alias Operations.{User, Backoffice, Notifications} operation User.Create operation Backoffice.SaveStats operation Notifications.SendEmail # or you can use step alias: # step User.Create # step Backoffice.SaveStats # step Notifications.SendEmail end This is how invoke this chain: iex&gt; CreateUser.run(name: &quot;User Name&quot;, age: 37, gender: &quot;m&quot;) Exop.Chain defines run/1 function under the hood (like common operations do) that accepts keyword(), map() or struct() as params. Those params will be passed into the first operation in the chain. Bear in mind that each of chained operations (except the first one) awaits a returned result of a previous operation as incoming params. So in the example above CreateUser.run(name: &quot;User Name&quot;, age: 37, gender: &quot;m&quot;) will invoke the chain by passing [name: &quot;User Name&quot;, age: 37, gender: &quot;m&quot;] params to the first User.Create operation. The result of User.Create operation will be passed to Backoffice.SaveStats operation as its params and so on. Once any of operations in the chain returns non-ok-tuple result (error result, interruption, auth error etc.) the chain execution interrupts and error result returned (as the chain (CreateUser) result). You can pass additional parameters to any operation in a chain (with either an exact value or 0-arity function): defmodule CreateUser do use Exop.Chain alias Operations.{User, Backoffice, Notifications} operation User.Create operation Backoffice.SaveStats, logger: MyFancyLoggerModule # or operation Backoffice.SaveStats, logger: &amp;__MODULE__.logger/0 operation Notifications.SendEmail def logger, do: MyFancyLoggerModule end use Exop.Chain can take :name_in_error option, when it is set to true a failed operation in a chain returns the operation&#39;s module name as the first elements of output tuple {YourOperation, {:error, _}} defmodule YourChain do use Exop.Chain, name_in_error: true operation Op1 operation Op2Fail operation Op3 end iex&gt; YourChain.run(a: &quot;1&quot;, b: 2) {Op2Fail, {:error, {:validation, %{a: [&quot;has wrong type&quot;]}}}} name_in_error: true doesn&#39;t affect operations with a fallback defined (unmodified fallback result is returned)."},{"ref":"readme.html#license","title":"Exop - LICENSE","type":"extras","doc":"Copyright © 2016 - 2019 Andrey Chernykh ( andrei.chernykh@gmail.com ) This work is free. You can redistribute it and/or modify it under the terms of the MIT License. See the LICENSE file for more details."}]